@@ -21,40 +21,41 @@ import (
 
 	"golang.org/x/tools/go/analysis"
 	"golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	goplsastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	"golang.org/x/tools/internal/analysisinternal"
-	"golang.org/x/tools/internal/astutil/cursor"
 	"golang.org/x/tools/internal/typesinternal"
 )
 
-// extractVariable implements the refactor.extract.{variable,constant} CodeAction command.
-func extractVariable(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
-	return extractExprs(pkg, pgf, start, end, false)
+// extractVariableOne implements the refactor.extract.{variable,constant} CodeAction command.
+func extractVariableOne(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
+	countExtractVariable.Inc()
+	return extractVariable(pkg, pgf, start, end, false)
 }
 
 // extractVariableAll implements the refactor.extract.{variable,constant}-all CodeAction command.
 func extractVariableAll(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
-	return extractExprs(pkg, pgf, start, end, true)
+	countExtractVariableAll.Inc()
+	return extractVariable(pkg, pgf, start, end, true)
 }
 
-// extractExprs replaces occurrence(s) of a specified expression within the same function
-// with newVar. If 'all' is true, it replaces all occurrences of the same expression;
-// otherwise, it only replaces the selected expression.
+// extractVariable replaces one or all occurrences of a specified
+// expression within the same function with newVar. If 'all' is true,
+// it replaces all occurrences of the same expression; otherwise, it
+// only replaces the selected expression.
 //
 // The new variable/constant is declared as close as possible to the first found expression
 // within the deepest common scope accessible to all candidate occurrences.
-func extractExprs(pkg *cache.Package, pgf *parsego.File, start, end token.Pos, all bool) (*token.FileSet, *analysis.SuggestedFix, error) {
+func extractVariable(pkg *cache.Package, pgf *parsego.File, start, end token.Pos, all bool) (*token.FileSet, *analysis.SuggestedFix, error) {
 	var (
 		fset = pkg.FileSet()
 		info = pkg.TypesInfo()
 		file = pgf.File
 	)
 	// TODO(adonovan): simplify, using Cursor.
-	tokFile := fset.File(file.FileStart)
 	exprs, err := canExtractVariable(info, pgf.Cursor, start, end, all)
 	if err != nil {
 		return nil, nil, fmt.Errorf("cannot extract: %v", err)
@@ -164,7 +165,7 @@ Outer:
 			return nil, nil, fmt.Errorf("cannot find location to insert extraction: %v", err)
 		}
 		// Within function: compute appropriate statement indentation.
-		indent, err := calculateIndentation(pgf.Src, tokFile, before)
+		indent, err := pgf.Indentation(before.Pos())
 		if err != nil {
 			return nil, nil, err
 		}
@@ -391,7 +392,7 @@ func stmtToInsertVarBefore(path []ast.Node, variables []*variable) (ast.Stmt, er
 // canExtractVariable reports whether the code in the given range can be
 // extracted to a variable (or constant). It returns the selected expression or, if 'all',
 // all structurally equivalent expressions within the same function body, in lexical order.
-func canExtractVariable(info *types.Info, curFile cursor.Cursor, start, end token.Pos, all bool) ([]ast.Expr, error) {
+func canExtractVariable(info *types.Info, curFile inspector.Cursor, start, end token.Pos, all bool) ([]ast.Expr, error) {
 	if start == end {
 		return nil, fmt.Errorf("empty selection")
 	}
@@ -506,19 +507,6 @@ func canExtractVariable(info *types.Info, curFile cursor.Cursor, start, end toke
 	return exprs, nil
 }
 
-// Calculate indentation for insertion.
-// When inserting lines of code, we must ensure that the lines have consistent
-// formatting (i.e. the proper indentation). To do so, we observe the indentation on the
-// line of code on which the insertion occurs.
-func calculateIndentation(content []byte, tok *token.File, insertBeforeStmt ast.Node) (string, error) {
-	line := safetoken.Line(tok, insertBeforeStmt.Pos())
-	lineOffset, stmtOffset, err := safetoken.Offsets(tok, tok.LineStart(line), insertBeforeStmt.Pos())
-	if err != nil {
-		return "", err
-	}
-	return string(content[lineOffset:stmtOffset]), nil
-}
-
 // freshName returns an identifier based on prefix (perhaps with a
 // numeric suffix) that is not in scope at the specified position
 // within the file. It returns the next numeric suffix to use.
@@ -582,11 +570,13 @@ type returnVariable struct {
 
 // extractMethod refactors the selected block of code into a new method.
 func extractMethod(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
+	countExtractMethod.Inc()
 	return extractFunctionMethod(pkg, pgf, start, end, true)
 }
 
 // extractFunction refactors the selected block of code into a new function.
 func extractFunction(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
+	countExtractFunction.Inc()
 	return extractFunctionMethod(pkg, pgf, start, end, false)
 }
 
@@ -627,31 +617,75 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 	// A return statement is non-nested if its parent node is equal to the parent node
 	// of the first node in the selection. These cases must be handled separately because
 	// non-nested return statements are guaranteed to execute.
-	var retStmts []*ast.ReturnStmt
 	var hasNonNestedReturn bool
-	startParent := findParent(outer, node)
-	ast.Inspect(outer, func(n ast.Node) bool {
-		if n == nil {
-			return false
+	curStart, ok := pgf.Cursor.FindNode(node)
+	if !ok {
+		return nil, nil, bug.Errorf("cannot find Cursor for start Node")
+	}
+	curOuter, ok := pgf.Cursor.FindNode(outer)
+	if !ok {
+		return nil, nil, bug.Errorf("cannot find Cursor for start Node")
+	}
+	// Determine whether all return statements in the selection are
+	// error-handling return statements. They must be of the form:
+	// if err != nil {
+	// 	return ..., err
+	// }
+	// If all return statements in the extracted block have a non-nil error, we
+	// can replace the "shouldReturn" check with an error check to produce a
+	// more concise output.
+	allReturnsFinalErr := true // all ReturnStmts have final 'err' expression
+	hasReturn := false         // selection contains a ReturnStmt
+	filter := []ast.Node{(*ast.ReturnStmt)(nil), (*ast.FuncLit)(nil)}
+	curOuter.Inspect(filter, func(cur inspector.Cursor) (descend bool) {
+		if funcLit, ok := cur.Node().(*ast.FuncLit); ok {
+			// Exclude return statements in function literals because they don't affect the refactor.
+			// Keep descending into func lits whose declaration is not included in the extracted block.
+			return !(start < funcLit.Pos() && funcLit.End() < end)
 		}
-		if n.Pos() < start || n.End() > end {
-			return n.Pos() <= end
+		ret := cur.Node().(*ast.ReturnStmt)
+		if ret.Pos() < start || ret.End() > end {
+			return false // not part of the extracted block
 		}
-		// exclude return statements in function literals because they don't affect the refactor.
-		if _, ok := n.(*ast.FuncLit); ok {
-			return false
-		}
-		ret, ok := n.(*ast.ReturnStmt)
-		if !ok {
-			return true
-		}
-		if findParent(outer, n) == startParent {
+		hasReturn = true
+
+		if cur.Parent() == curStart.Parent() {
 			hasNonNestedReturn = true
 		}
-		retStmts = append(retStmts, ret)
+
+		if !allReturnsFinalErr {
+			// Stop the traversal if we have already found a non error-handling return statement.
+			return false
+		}
+		// Check if the return statement returns a non-nil error as the last value.
+		if len(ret.Results) > 0 {
+			typ := info.TypeOf(ret.Results[len(ret.Results)-1])
+			if typ != nil && types.Identical(typ, errorType) {
+				// Have: return ..., err
+				// Check for enclosing "if err != nil { return ..., err }".
+				// In that case, we can lift the error return to the caller.
+				if ifstmt, ok := cur.Parent().Parent().Node().(*ast.IfStmt); ok {
+					// Only handle the case where the if statement body contains a single statement.
+					if body, ok := cur.Parent().Node().(*ast.BlockStmt); ok && len(body.List) <= 1 {
+						if cond, ok := ifstmt.Cond.(*ast.BinaryExpr); ok {
+							tx := info.TypeOf(cond.X)
+							ty := info.TypeOf(cond.Y)
+							isErr := tx != nil && types.Identical(tx, errorType)
+							isNil := ty != nil && types.Identical(ty, types.Typ[types.UntypedNil])
+							if cond.Op == token.NEQ && isErr && isNil {
+								// allReturnsErrHandling remains true
+								return false
+							}
+						}
+					}
+				}
+			}
+		}
+		allReturnsFinalErr = false
 		return false
 	})
-	containsReturnStatement := len(retStmts) > 0
+
+	allReturnsFinalErr = hasReturn && allReturnsFinalErr
 
 	// Now that we have determined the correct range for the selection block,
 	// we must determine the signature of the extracted function. We will then replace
@@ -738,7 +772,9 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		// cannot be its own reassignment or redefinition (objOverriden).
 		vscope := v.obj.Parent()
 		if vscope == nil {
-			return nil, nil, fmt.Errorf("parent nil")
+			// v.obj could be a field on an anonymous struct. We'll examine the
+			// struct in a different iteration so don't return an error here.
+			continue
 		}
 		isUsed, firstUseAfter := objUsed(info, end, vscope.End(), v.obj)
 		if v.assigned && isUsed && !varOverridden(info, firstUseAfter, v.obj, v.free, outer) {
@@ -753,6 +789,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 				//
 				// The second condition below handles the case when
 				// v's block is the FuncDecl.Body itself.
+				startParent := curStart.Parent().Node()
 				if vscope.Pos() == startParent.Pos() ||
 					startParent == outer.Body && vscope == info.Scopes[outer.Type] {
 					canRedefineCount++
@@ -893,13 +930,26 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 
 	var retVars []*returnVariable
 	var ifReturn *ast.IfStmt
-	if containsReturnStatement {
+
+	// Determine if the extracted block contains any free branch statements, for
+	// example: "continue label" where "label" is declared outside of the
+	// extracted block, or continue inside a "for" statement where the for
+	// statement is declared outside of the extracted block. These will be
+	// handled below, after adjusting return statements and generating return
+	// info.
+	curSel, _ := pgf.Cursor.FindByPos(start, end) // since canExtractFunction succeeded, this will always return a valid cursor
+	freeBranches := freeBranches(info, curSel, start, end)
+
+	// All return statements in the extracted block are error handling returns, and there are no free control statements.
+	isErrHandlingReturnsCase := allReturnsFinalErr && len(freeBranches) == 0
+
+	if hasReturn {
 		if !hasNonNestedReturn {
 			// The selected block contained return statements, so we have to modify the
 			// signature of the extracted function as described above. Adjust all of
 			// the return statements in the extracted function to reflect this change in
 			// signature.
-			if err := adjustReturnStatements(returnTypes, seenVars, extractedBlock, qual); err != nil {
+			if err := adjustReturnStatements(returnTypes, seenVars, extractedBlock, qual, isErrHandlingReturnsCase); err != nil {
 				return nil, nil, err
 			}
 		}
@@ -907,17 +957,12 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		// statements in the selection. Update the type signature of the extracted
 		// function and construct the if statement that will be inserted in the enclosing
 		// function.
-		retVars, ifReturn, err = generateReturnInfo(enclosing, pkg, path, file, info, start, end, hasNonNestedReturn)
+		retVars, ifReturn, err = generateReturnInfo(enclosing, pkg, path, file, info, start, end, hasNonNestedReturn, isErrHandlingReturnsCase)
 		if err != nil {
 			return nil, nil, err
 		}
 	}
 
-	// Determine if the extracted block contains any free branch statements, for
-	// example: "continue label" where "label" is declared outside of the
-	// extracted block, or continue inside a "for" statement where the for
-	// statement is declared outside of the extracted block.
-
 	// If the extracted block contains free branch statements, we add another
 	// return value "ctrl" to the extracted function that will be used to
 	// determine the control flow. See the following example, where === denotes
@@ -955,9 +1000,6 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 	// }
 	//
 
-	curSel, _ := pgf.Cursor.FindByPos(start, end) // since canExtractFunction succeeded, this will always return a valid cursor
-	freeBranches := freeBranches(info, curSel, start, end)
-
 	// Generate an unused identifier for the control value.
 	ctrlVar, _ := freshName(info, file, start, "ctrl", 0)
 	if len(freeBranches) > 0 {
@@ -1078,8 +1120,16 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		return nil, nil, err
 	}
 	if ifReturn != nil {
-		if err := format.Node(&ifBuf, fset, ifReturn); err != nil {
-			return nil, nil, err
+		if isErrHandlingReturnsCase {
+			errName := retVars[len(retVars)-1]
+			fmt.Fprintf(&ifBuf, "if %s != nil ", errName.name.String())
+			if err := format.Node(&ifBuf, fset, ifReturn.Body); err != nil {
+				return nil, nil, err
+			}
+		} else {
+			if err := format.Node(&ifBuf, fset, ifReturn); err != nil {
+				return nil, nil, err
+			}
 		}
 	}
 
@@ -1135,7 +1185,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 	}
 	before := src[outerStart:startOffset]
 	after := src[endOffset:outerEnd]
-	indent, err := calculateIndentation(src, tok, node)
+	indent, err := pgf.Indentation(node.Pos())
 	if err != nil {
 		return nil, nil, err
 	}
@@ -1165,11 +1215,11 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		for i, br := range freeBranches {
 			// Preserve spacing at the beginning of the line containing the branch statement.
 			startPos := tok.LineStart(safetoken.Line(tok, br.Pos()))
-			start, end, err := safetoken.Offsets(tok, startPos, br.End())
+			text, err := pgf.PosText(startPos, br.End())
 			if err != nil {
 				return nil, nil, err
 			}
-			fmt.Fprintf(&fullReplacement, "case %d:\n%s%s", i+1, pgf.Src[start:end], newLineIndent)
+			fmt.Fprintf(&fullReplacement, "case %d:\n%s%s", i+1, text, newLineIndent)
 		}
 		fullReplacement.WriteString("}")
 	}
@@ -1238,7 +1288,7 @@ func moveParamToFrontIfFound(params []ast.Expr, paramTypes []*ast.Field, x, sel
 // their cursors for whitespace. To support this use case, we must manually adjust the
 // ranges to match the correct AST node. In this particular example, we would adjust
 // rng.Start forward to the start of 'if' and rng.End backward to after '}'.
-func adjustRangeForCommentsAndWhiteSpace(tok *token.File, start, end token.Pos, content []byte, curFile cursor.Cursor) (token.Pos, token.Pos, error) {
+func adjustRangeForCommentsAndWhiteSpace(tok *token.File, start, end token.Pos, content []byte, curFile inspector.Cursor) (token.Pos, token.Pos, error) {
 	file := curFile.Node().(*ast.File)
 	// TODO(adonovan): simplify, using Cursor.
 
@@ -1306,20 +1356,6 @@ func isGoWhiteSpace(b byte) bool {
 	return uint64(scanner.GoWhitespace)&(1<<uint(b)) != 0
 }
 
-// findParent finds the parent AST node of the given target node, if the target is a
-// descendant of the starting node.
-func findParent(start ast.Node, target ast.Node) ast.Node {
-	var parent ast.Node
-	analysisinternal.WalkASTWithParent(start, func(n, p ast.Node) bool {
-		if n == target {
-			parent = p
-			return false
-		}
-		return true
-	})
-	return parent
-}
-
 // variable describes the status of a variable within a selection.
 type variable struct {
 	obj types.Object
@@ -1532,6 +1568,12 @@ func collectFreeVars(info *types.Info, file *ast.File, start, end token.Pos, nod
 		if !ok {
 			return nil, fmt.Errorf("no seen types.Object for %v", obj)
 		}
+		if named, ok := v.obj.Type().(typesinternal.NamedOrAlias); ok {
+			namedPos := named.Obj().Pos()
+			if isLocal(named.Obj()) && !(start <= namedPos && namedPos <= end) {
+				return nil, fmt.Errorf("Cannot extract selection: the code refers to a local type whose definition lies outside the extracted block")
+			}
+		}
 		variables = append(variables, v)
 	}
 	return variables, nil
@@ -1568,7 +1610,7 @@ type fnExtractParams struct {
 
 // canExtractFunction reports whether the code in the given range can be
 // extracted to a function.
-func canExtractFunction(tok *token.File, start, end token.Pos, src []byte, curFile cursor.Cursor) (*fnExtractParams, bool, bool, error) {
+func canExtractFunction(tok *token.File, start, end token.Pos, src []byte, curFile inspector.Cursor) (*fnExtractParams, bool, bool, error) {
 	if start == end {
 		return nil, false, false, fmt.Errorf("start and end are equal")
 	}
@@ -1734,19 +1776,9 @@ func parseStmts(fset *token.FileSet, src []byte) (*ast.BlockStmt, []*ast.Comment
 // signature of the extracted function. We prepare names, signatures, and "zero values" that
 // represent the new variables. We also use this information to construct the if statement that
 // is inserted below the call to the extracted function.
-func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, path []ast.Node, file *ast.File, info *types.Info, start, end token.Pos, hasNonNestedReturns bool) ([]*returnVariable, *ast.IfStmt, error) {
+func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, path []ast.Node, file *ast.File, info *types.Info, start, end token.Pos, hasNonNestedReturns bool, isErrHandlingReturnsCase bool) ([]*returnVariable, *ast.IfStmt, error) {
 	var retVars []*returnVariable
 	var cond *ast.Ident
-	if !hasNonNestedReturns {
-		// Generate information for the added bool value.
-		name, _ := freshNameOutsideRange(info, file, path[0].Pos(), start, end, "shouldReturn", 0)
-		cond = &ast.Ident{Name: name}
-		retVars = append(retVars, &returnVariable{
-			name:    cond,
-			decl:    &ast.Field{Type: ast.NewIdent("bool")},
-			zeroVal: ast.NewIdent("false"),
-		})
-	}
 	// Generate information for the values in the return signature of the enclosing function.
 	if enclosing.Results != nil {
 		nameIdx := make(map[string]int) // last integral suffixes of generated names
@@ -1787,12 +1819,21 @@ func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, path []ast.
 	}
 	var ifReturn *ast.IfStmt
 	if !hasNonNestedReturns {
-		// Create the return statement for the enclosing function. We must exclude the variable
-		// for the condition of the if statement (cond) from the return statement.
+		results := getNames(retVars)
+		if !isErrHandlingReturnsCase {
+			// Generate information for the added bool value.
+			name, _ := freshNameOutsideRange(info, file, path[0].Pos(), start, end, "shouldReturn", 0)
+			cond = &ast.Ident{Name: name}
+			retVars = append(retVars, &returnVariable{
+				name:    cond,
+				decl:    &ast.Field{Type: ast.NewIdent("bool")},
+				zeroVal: ast.NewIdent("false"),
+			})
+		}
 		ifReturn = &ast.IfStmt{
 			Cond: cond,
 			Body: &ast.BlockStmt{
-				List: []ast.Stmt{&ast.ReturnStmt{Results: getNames(retVars)[1:]}},
+				List: []ast.Stmt{&ast.ReturnStmt{Results: results}},
 			},
 		}
 	}
@@ -1819,31 +1860,26 @@ var conventionalVarNames = map[objKey]string{
 //
 // For special types, it uses known conventional names.
 func varNameForType(t types.Type) (string, bool) {
-	var typeName string
-	if tn, ok := t.(interface{ Obj() *types.TypeName }); ok {
-		obj := tn.Obj()
-		k := objKey{name: obj.Name()}
-		if obj.Pkg() != nil {
-			k.pkg = obj.Pkg().Name()
-		}
-		if name, ok := conventionalVarNames[k]; ok {
-			return name, true
-		}
-		typeName = obj.Name()
-	} else if b, ok := t.(*types.Basic); ok {
-		typeName = b.Name()
-	}
-
-	if typeName == "" {
+	tname := typesinternal.TypeNameFor(t)
+	if tname == nil {
 		return "", false
 	}
 
-	return AbbreviateVarName(typeName), true
+	// Have Alias, Basic, Named, or TypeParam.
+	k := objKey{name: tname.Name()}
+	if tname.Pkg() != nil {
+		k.pkg = tname.Pkg().Name()
+	}
+	if name, ok := conventionalVarNames[k]; ok {
+		return name, true
+	}
+
+	return AbbreviateVarName(tname.Name()), true
 }
 
 // adjustReturnStatements adds "zero values" of the given types to each return
 // statement in the given AST node.
-func adjustReturnStatements(returnTypes []*ast.Field, seenVars map[types.Object]ast.Expr, extractedBlock *ast.BlockStmt, qual types.Qualifier) error {
+func adjustReturnStatements(returnTypes []*ast.Field, seenVars map[types.Object]ast.Expr, extractedBlock *ast.BlockStmt, qual types.Qualifier, isErrHandlingReturnsCase bool) error {
 	var zeroVals []ast.Expr
 	// Create "zero values" for each type.
 	for _, returnType := range returnTypes {
@@ -1865,11 +1901,17 @@ func adjustReturnStatements(returnTypes []*ast.Field, seenVars map[types.Object]
 	// extracted function. We set the bool to 'true' because, if these return statements
 	// execute, the extracted function terminates early, and the enclosing function must
 	// return as well.
-	zeroVals = append(zeroVals, ast.NewIdent("true"))
+	if !isErrHandlingReturnsCase {
+		zeroVals = append(zeroVals, ast.NewIdent("true"))
+	}
 	ast.Inspect(extractedBlock, func(n ast.Node) bool {
 		if n == nil {
 			return false
 		}
+		// Don't modify return statements inside anonymous functions.
+		if _, ok := n.(*ast.FuncLit); ok {
+			return false
+		}
 		if n, ok := n.(*ast.ReturnStmt); ok {
 			n.Results = slices.Concat(zeroVals, n.Results)
 			return false
@@ -2022,7 +2064,7 @@ func replaceBranchStmtWithReturnStmt(block ast.Node, br *ast.BranchStmt, ret *as
 
 // freeBranches returns all branch statements beneath cur whose continuation
 // lies outside the (start, end) range.
-func freeBranches(info *types.Info, cur cursor.Cursor, start, end token.Pos) (free []*ast.BranchStmt) {
+func freeBranches(info *types.Info, cur inspector.Cursor, start, end token.Pos) (free []*ast.BranchStmt) {
 nextBranch:
 	for curBr := range cur.Preorder((*ast.BranchStmt)(nil)) {
 		br := curBr.Node().(*ast.BranchStmt)
