@@ -38,13 +38,13 @@ import (
 	"golang.org/x/tools/gopls/internal/protocol/command"
 	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/util/bug"
+	"golang.org/x/tools/gopls/internal/util/tokeninternal"
 	"golang.org/x/tools/gopls/internal/vulncheck"
 	"golang.org/x/tools/gopls/internal/vulncheck/scan"
 	"golang.org/x/tools/internal/diff"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/gocommand"
 	"golang.org/x/tools/internal/jsonrpc2"
-	"golang.org/x/tools/internal/tokeninternal"
 	"golang.org/x/tools/internal/xcontext"
 )
 
@@ -91,7 +91,7 @@ func (h *commandHandler) Modules(ctx context.Context, args command.ModulesArgs)
 			return false // "can't happen" (see prior Encloses check)
 		}
 
-		assert(filepath.Base(goMod.Path()) == "go.mod", fmt.Sprintf("invalid go.mod path: want go.mod, got %q", goMod.Path()))
+		assert(goMod.Base() == "go.mod", fmt.Sprintf("invalid go.mod path: want go.mod, got %q", goMod.Path()))
 
 		// Invariant: rel is a relative path without "../" segments and the last
 		// segment is "go.mod"
@@ -292,13 +292,19 @@ func (c *commandHandler) AddTest(ctx context.Context, loc protocol.Location) (*p
 		if deps.snapshot.FileKind(deps.fh) != file.Go {
 			return fmt.Errorf("can't add test for non-Go file")
 		}
-		docedits, err := golang.AddTestForFunc(ctx, deps.snapshot, loc)
+		docedits, show, err := golang.AddTestForFunc(ctx, deps.snapshot, loc)
 		if err != nil {
 			return err
 		}
-		return applyChanges(ctx, c.s.client, docedits)
+		if err := applyChanges(ctx, c.s.client, docedits); err != nil {
+			return err
+		}
+
+		if show != nil {
+			showDocumentImpl(ctx, c.s.client, protocol.URI(show.URI), &show.Range, c.s.options)
+		}
+		return nil
 	})
-	// TODO(hxjiang): move the cursor to the new test once edits applied.
 	return result, err
 }
 
@@ -355,7 +361,7 @@ func (c *commandHandler) run(ctx context.Context, cfg commandConfig, run command
 		return bug.Errorf("internal error: forURI=%q, forView=%q", cfg.forURI, cfg.forView)
 	}
 	if cfg.forURI != "" {
-		deps.fh, deps.snapshot, release, err = c.s.fileOf(ctx, cfg.forURI)
+		deps.fh, deps.snapshot, release, err = c.s.session.FileOf(ctx, cfg.forURI)
 		if err != nil {
 			return err
 		}
@@ -546,7 +552,7 @@ func (c *commandHandler) UpdateGoSum(ctx context.Context, args command.URIArgs)
 		progress: "Updating go.sum",
 	}, func(ctx context.Context, _ commandDeps) error {
 		for _, uri := range args.URIs {
-			fh, snapshot, release, err := c.s.fileOf(ctx, uri)
+			fh, snapshot, release, err := c.s.session.FileOf(ctx, uri)
 			if err != nil {
 				return err
 			}
@@ -567,7 +573,7 @@ func (c *commandHandler) Tidy(ctx context.Context, args command.URIArgs) error {
 		progress: "Running go mod tidy",
 	}, func(ctx context.Context, _ commandDeps) error {
 		for _, uri := range args.URIs {
-			fh, snapshot, release, err := c.s.fileOf(ctx, uri)
+			fh, snapshot, release, err := c.s.session.FileOf(ctx, uri)
 			if err != nil {
 				return err
 			}
@@ -616,7 +622,7 @@ func (c *commandHandler) EditGoDirective(ctx context.Context, args command.EditG
 		requireSave: true, // if go.mod isn't saved it could cause a problem
 		forURI:      args.URI,
 	}, func(ctx context.Context, _ commandDeps) error {
-		fh, snapshot, release, err := c.s.fileOf(ctx, args.URI)
+		fh, snapshot, release, err := c.s.session.FileOf(ctx, args.URI)
 		if err != nil {
 			return err
 		}
@@ -709,7 +715,7 @@ func (c *commandHandler) Doc(ctx context.Context, args command.DocArgs) (protoco
 		}
 
 		// Compute package path and optional symbol fragment
-		// (e.g. "#Buffer.Len") from the the selection.
+		// (e.g. "#Buffer.Len") from the selection.
 		pkgpath, fragment, _ := golang.DocFragment(pkg, pgf, start, end)
 
 		// Direct the client to open the /pkg page.
@@ -736,7 +742,7 @@ func (c *commandHandler) RunTests(ctx context.Context, args command.RunTestsArgs
 
 func (c *commandHandler) runTests(ctx context.Context, snapshot *cache.Snapshot, work *progress.WorkDone, uri protocol.DocumentURI, tests, benchmarks []string) error {
 	// TODO: fix the error reporting when this runs async.
-	meta, err := golang.NarrowestMetadataForFile(ctx, snapshot, uri)
+	meta, err := snapshot.NarrowestMetadataForFile(ctx, uri)
 	if err != nil {
 		return err
 	}
@@ -1021,7 +1027,7 @@ func (c *commandHandler) GCDetails(ctx context.Context, uri protocol.DocumentURI
 	}, func(ctx context.Context, deps commandDeps) error {
 		return c.modifyState(ctx, FromToggleCompilerOptDetails, func() (*cache.Snapshot, func(), error) {
 			// Don't blindly use "dir := deps.fh.URI().Dir()"; validate.
-			meta, err := golang.NarrowestMetadataForFile(ctx, deps.snapshot, deps.fh.URI())
+			meta, err := deps.snapshot.NarrowestMetadataForFile(ctx, deps.fh.URI())
 			if err != nil {
 				return nil, nil, err
 			}
@@ -1082,7 +1088,7 @@ func (c *commandHandler) ListImports(ctx context.Context, args command.URIArg) (
 				})
 			}
 		}
-		meta, err := golang.NarrowestMetadataForFile(ctx, deps.snapshot, args.URI)
+		meta, err := deps.snapshot.NarrowestMetadataForFile(ctx, args.URI)
 		if err != nil {
 			return err // e.g. cancelled
 		}
@@ -1551,7 +1557,7 @@ func openClientBrowser(ctx context.Context, cli protocol.Client, title string, u
 			Message: fmt.Sprintf("%s: open your browser to %s", title, url),
 		}
 		if err := cli.ShowMessage(ctx, params); err != nil {
-			event.Error(ctx, "failed to show brower url", err)
+			event.Error(ctx, "failed to show browser url", err)
 		}
 	}
 }
@@ -1604,6 +1610,7 @@ func showDocumentImpl(ctx context.Context, cli protocol.Client, url protocol.URI
 }
 
 func (c *commandHandler) ChangeSignature(ctx context.Context, args command.ChangeSignatureArgs) (*protocol.WorkspaceEdit, error) {
+	countChangeSignature.Inc()
 	var result *protocol.WorkspaceEdit
 	err := c.run(ctx, commandConfig{
 		forURI: args.Location.URI,
@@ -1650,7 +1657,7 @@ func (c *commandHandler) DiagnoseFiles(ctx context.Context, args command.Diagnos
 
 		snapshots := make(map[*cache.Snapshot]bool)
 		for _, uri := range args.Files {
-			fh, snapshot, release, err := c.s.fileOf(ctx, uri)
+			fh, snapshot, release, err := c.s.session.FileOf(ctx, uri)
 			if err != nil {
 				return err
 			}
@@ -1704,6 +1711,16 @@ func (c *commandHandler) FreeSymbols(ctx context.Context, viewID string, loc pro
 	return nil
 }
 
+func (c *commandHandler) SplitPackage(ctx context.Context, viewID, packageID string) error {
+	web, err := c.s.getWeb()
+	if err != nil {
+		return err
+	}
+	url := web.splitpkgURL(viewID, packageID)
+	openClientBrowser(ctx, c.s.client, "SplitPackage", url, c.s.Options())
+	return nil
+}
+
 func (c *commandHandler) Assembly(ctx context.Context, viewID, packageID, symbol string) error {
 	web, err := c.s.getWeb()
 	if err != nil {
@@ -1789,8 +1806,13 @@ func (c *commandHandler) ModifyTags(ctx context.Context, args command.ModifyTags
 		}
 		m.Transform = transform
 
+		// Each command involves either adding or removing tags, depending on
+		// whether Add or Clear is set.
 		if args.Add != "" {
+			countAddStructTags.Inc()
 			m.Add = strings.Split(args.Add, ",")
+		} else if args.Clear {
+			countRemoveStructTags.Inc()
 		}
 		if args.AddOptions != "" {
 			if options, err := optionsStringToMap(args.AddOptions); err != nil {
